"""
.. module: onacol.config_schema
   :synopsis: Classes and utilities for parsing and manipulation of the
                configuration schema.

.. moduleauthor:: Josef Nevrly <josef.nevrly@gmail.com>
"""
from typing import Any, List, Union
import copy

from cerberus.schema import SchemaRegistry  # type: ignore

from .base import OnacolException
from .flat_schema import FlatValueType, FlatSchemaMetadata


class SchemaException(OnacolException):
    pass


def _has_subelement(source_element: Any, subelement: Any) -> bool:
    """ Fail-safe check if sub-element is part of the source element.
    """
    try:
        return (subelement in source_element)
    except TypeError:
        return False


# def _get_subelement(source_element: Any,
#                     subelement: Any,
#                     default: Any = None) -> Any:
#     """ Get sub-element from the source element. If not possible, gracefully
#         return default value.
#     """
#     try:
#         return source_element.get(subelement, default)
#     except AttributeError:
#         return default


class ConfigSchema:

    OC_SCHEMA = "oc_schema"
    OC_SCHEMA_ID = "oc_schema_id"
    OC_DESC = "oc_desc"
    OC_DEFAULT = "oc_default"

    # All yaml tokens used for onacol schema metadata.
    # Note: OC_DESC is to collect descriptions. This feature is not used in the
    # current API, but is kept there in case...
    OC_TOKENS = [OC_SCHEMA, OC_SCHEMA_ID, OC_DESC, OC_DEFAULT]

    def __init__(self, schema_source: dict):
        """
        :param schema_source: Configuration schema dictionary.
        """
        self._schema_source: dict = schema_source
        self._schema: dict = {}
        self._flat_schema: dict = {}  # Used for ENV_VAR list
        self._defaults: dict = {}
        self._descriptions: dict = {}
        self._validator = None
        self._schema_registry = SchemaRegistry()
        self.parse_schema(self._schema_source)

    def __bool__(self):
        return bool(self._schema)

    @property
    def schema(self) -> dict:
        """ Configuration schema.
        """
        return self._schema

    @property
    def flat_schema(self) -> dict:
        """ Flattened configuration schema
            (used for mapping environment variables).
        """
        return self._flat_schema

    @property
    def schema_registry(self) -> SchemaRegistry:
        """ Cerberus SchemaRegistry object used for configuration validation.
        """
        return self._schema_registry

    @property
    def defaults(self) -> dict:
        """ Configuration default values. """
        return self._defaults

    # @property
    # def descriptions(self) -> dict:
    #     """ Config item descriptions (not used in current version.)"""
    #     return self._descriptions

    def _element_is_leaf(self, source_element: Any) -> bool:
        """ Check if current configuration element is a leaf in terms of
            configuration tree traversal (i.e. it has no deeper levels).
        """
        # Leaf element means it either just has a value, or it contains
        # dict that is purely onacol related metadata
        if isinstance(source_element, dict):
            # Check if it has only onacol metadata
            for k in source_element.keys():
                if k not in self.OC_TOKENS:
                    return False
            return True
        elif isinstance(source_element, list):
            return False
        else:
            # It's just a value
            return True

    def parse_schema(self, schema_source: dict):
        """ Parse default configuration including schema metadata.
            During parsing, the schema for validation, default configuration
            values and flattened schema for environment variable parsing is
            processed.
        """
        self._flat_schema = {}

        # No meaning parsing empty schema
        if not schema_source:
            return

        self._schema, self._defaults, self._descriptions = \
            self._process_schema_element(schema_source, [], top_level=True)

    def _process_schema_element(self, schema_source: Any,
                                document_path: Union[List[str], None],
                                top_level: bool=False) -> tuple:
        """ Recursively parse given configuration element.

        :param schema_source:  Element of the configuration schema.
        :param document_path:  List of keys marking element's path
                                in the configuration hierarchy or None for
                                processing leaf elements.
        :param top_level:      Flag for identifying top-level element
                                (root of the configuration tree).
        :return:  Tuple (schema, defaults, description) with parsed elements of
                    schema, default values and element descriptions.
                    Element descriptions are not used in the current version
                    of the library.
        """
        schema = None  # type: ignore
        default = None
        description = None
        if self._element_is_leaf(schema_source):
            if _has_subelement(schema_source, self.OC_SCHEMA):
                schema = schema_source[self.OC_SCHEMA]
            else:
                schema = None

            # Register this path as possible env_var vector
            if document_path is not None:
                self._flat_schema[tuple(document_path)] = FlatSchemaMetadata(
                    FlatValueType.VALUE,
                    schema.get("type") if schema else None
                )

            # Process default value
            try:
                default = schema_source[self.OC_DEFAULT]
            except (KeyError, TypeError):
                default = schema_source

            # If description is present, put it in the description map
            try:
                description = schema_source[self.OC_DESC]
            except (KeyError, TypeError):
                description = None
        else:
            # It's just a branching to deeper dict or list
            if isinstance(schema_source, dict):

                # Process dict
                schema = {"type": "dict", "schema": {}}
                default = {}
                description = {}
                for k, v in schema_source.items():
                    if k in self.OC_TOKENS:
                        continue

                    # We will be doing further branching
                    if document_path is not None:
                        d_path = document_path.copy()
                        d_path.append(k)
                    else:
                        d_path = None   # type: ignore

                    _schema, _default, _description = \
                        self._process_schema_element(
                            v, d_path)

                    if _schema is not None:
                        schema["schema"][k] = _schema
                    default[k] = _default
                    description[k] = _description

            elif isinstance(schema_source, list):
                # Lists cannot be expanded as env_var names, so just
                # register this path as possible env_var vector
                if document_path is not None:
                    self._flat_schema[tuple(document_path)] = \
                        FlatSchemaMetadata(FlatValueType.LIST, None)

                # Process list
                schema = {"type": "list", "schema": {}}
                default = []
                description = []
                i = 0
                for item in schema_source:
                    _schema, _default, _description = \
                        self._process_schema_element(item, None)
                    if (i == 0) and (_schema is not None):
                        schema["schema"] = _schema
                    default.append(_default)
                    description.append(description)
                    i += 1

            if _has_subelement(schema_source, self.OC_SCHEMA):
                if isinstance(schema_source[self.OC_SCHEMA], dict):
                    schema.update(schema_source[self.OC_SCHEMA])  # type: ignore
                else:
                    schema["schema"] = schema_source[self.OC_SCHEMA]  # type: ignore

            if _has_subelement(schema_source, self.OC_SCHEMA_ID):
                if schema_source[self.OC_SCHEMA_ID] == schema["schema"]:  # type: ignore
                    raise SchemaException(
                        f"Schema self reference for {schema['schema']}")  # type: ignore
                if schema["schema"]:  # type: ignore
                    self._schema_registry.add(schema_source[self.OC_SCHEMA_ID],
                                              schema["schema"])  # type: ignore

        # For top-level element, remove the type & schema declaration
        if top_level:
            schema = schema["schema"]  # type: ignore

        return schema, default, description

    def _export_schema_element(self, schema_element: Any,
                               config_data: Any) -> Any:
        """ Recursively parse through the schema and current config to generate
            YAML file that retains the format of the default config file
            (including comments etc.) but contains the current config data.

        :param schema_element:  Element of the configuration schema.
        :param config_data:     Element of the current configuration, mathcing
                                element of the configuration schema.
        :return: Element of the final config file export.
        """
        if self._element_is_leaf(schema_element):
            return config_data
        else:
            # It's just a branching to deeper dict or list
            pop_list = []
            if isinstance(schema_element, dict):
                for k, v in schema_element.items():
                    if k in self.OC_TOKENS:
                        pop_list.append(k)
                        continue

                    schema_element[k] = \
                        self._export_schema_element(v, config_data[k])

                for k in pop_list:
                    schema_element.pop(k)

            elif isinstance(schema_element, list):

                trim_length = 0
                for i, item in enumerate(schema_element):
                    try:
                        schema_element[i] = \
                            self._export_schema_element(item, config_data[i])
                    except IndexError:
                        trim_length += 1

                # Trim the end or append extra config
                if trim_length > 0:
                    for i in range(trim_length):
                        schema_element.pop()
                else:
                    len_schema = len(schema_element)
                    len_config = len(config_data)
                    if len_schema < len_config:
                        for i in range(len_schema, len_config):
                            schema_element.append(config_data[i])

            return schema_element

    def schema_to_yaml(self, schema_yaml, config):
        """ Strips original schema document of schema metadata, replacing them
            with config values.

        :param schema:  Schema dict (as generated by Ruamel YAML)
        :param config:  Configuration dict.
        :return:  Schema dict updated by the configuration dict and stripped of
                    the Onacol metadata.
        """

        temp = copy.deepcopy(schema_yaml)
        self._export_schema_element(temp, config)
        return temp





"""Top-level package for Onacol."""

__author__ = """Josef Nevrly"""
__email__ = 'josef.nevrly@gmail.com'
__version__ = '0.2.3'

from .onacol import ConfigManager, OnacolException, ConfigValidationError



"""
.. module: onacol.config_file
   :synopsis: Classes and utilities for loading and saving of the configuration
                files.

.. moduleauthor:: Josef Nevrly <josef.nevrly@gmail.com>
"""
from typing import Union, List, TextIO
import logging

from ruamel.yaml import YAML, YAMLError
from cascadict import CascaDict  # type: ignore

from .base import OnacolException
from .config_schema import ConfigSchema

YAML_ACCESS = YAML()

logger = logging.getLogger("onacol")


class ConfigFileException(OnacolException):
    pass


class ConfigFileHandler:

    def __init__(self, default_file_path: str,
                 optional_file_paths: Union[List[str], None] = None):
        """

        :param default_file_path:   Path with default configuration file that
                                    also defines the configuration schema.
        :param optional_file_paths: List of additional/optional configuration
                                    files that will be merged on top of the
                                    default configuration file.
        """
        self._default_file_path = default_file_path
        self._config = CascaDict({})
        self._schema: ConfigSchema = ConfigSchema({})
        self._schema_yaml: dict = {}  # Stored separately to preserve comments
        self._optional_file_paths = optional_file_paths or []
        self.load_files()

    @property
    def default_config_file(self) -> str:
        return self._default_file_path

    @property
    def default_config(self) -> dict:
        return self._schema.defaults

    @property
    def optional_config_files(self) -> List[str]:
        return self._optional_file_paths

    @property
    def configuration(self) -> dict:
        return self._config

    @configuration.setter
    def configuration(self, value: dict):
        self._config = value

    @property
    def config_schema(self) -> ConfigSchema:
        return self._schema

    @property
    def flat_schema(self) -> dict:
        return self._schema.flat_schema

    @property
    def has_defaults(self) -> bool:
        return self._default_file_path is not None

    @staticmethod
    def _load_yaml_file(yaml_file_path: str) -> dict:
        try:
            with open(yaml_file_path) as yaml_file:
                return YAML_ACCESS.load(yaml_file)
        except YAMLError as ye:
            raise ConfigFileException(f"Cannot parse config file: {str(ye)}")
        # except FileNotFoundError as fnf:
        #     raise ConfigFileException(f"Error reading file: {str(fnf)}")

    def load_files(self) -> None:
        """ Load default and optional config file and parse them into the
            configuration.
        """
        if self.has_defaults:
            self._schema_yaml = self._load_yaml_file(self._default_file_path)
            self._schema = ConfigSchema(self._schema_yaml)
            self._config = CascaDict(self._schema.defaults)
        else:
            self._config = CascaDict({})
            self._schema = ConfigSchema({})

        for opt_file in self._optional_file_paths:
            try:
                self.load_additional_file(opt_file)
            except FileNotFoundError:
                logger.warning("Optional config file at %s not found.",
                               opt_file)

    def load_additional_file(self, file_path):
        """ Load additional config file. If previous config is defined, it will
            be merged on top of the previous config.
        :param file_path:  Config file path.
        """
        if self._config:
            self._config = self._config.cascade(self._load_yaml_file(file_path))
        else:
            self._config = CascaDict(self._load_yaml_file(file_path))

    def save_with_schema(self, config: dict, save_file: TextIO) -> None:
        """ Save the configuration to the YAML file, keeping the original
            schema file form (including comments etc.).

        :param config: The configuration to be saved.
        :param save_file: Destination file-like (text) object.
        """
        # This dump converts None values to empty strings
        # (that is valid YAML 1.2)
        # Leaving as it is, if it becomes problem, here is a solution:
        # https://stackoverflow.com/a/44314840
        YAML_ACCESS.dump(
            self._schema.schema_to_yaml(self._schema_yaml, config),
            save_file
        )



"""
.. module: onacol.onacol
   :synopsis: Main module

.. moduleauthor:: Josef Nevrly <josef.nevrly@gmail.com>
"""
import os
from typing import List, TextIO, Any

from cerberus import Validator  # type: ignore
from cascadict import CascaDict  # type: ignore

from .config_file import ConfigFileHandler
from .flat_schema import FlatSchemaHandler

from .base import OnacolException


class ConfigValidationError(OnacolException):
    pass


class ConfigManager:

    def __init__(self, default_config_file_path: str,
                 optional_files: List[str] = None,
                 env_var_prefix: str = ""):
        """

        :param default_config_file_path: Path to the file with the default
                                         configuration with (optional)
                                         validation schema metadata.
        :param optional_files:   List of optional config files.
        :param env_var_prefix:   Prefix used for environment variables that
                                 shall be loaded as config.
        """
        self._file_handler = ConfigFileHandler(default_config_file_path,
                                               optional_files)
        self._flat_schema_handler = FlatSchemaHandler(
            self._file_handler.flat_schema, env_var_prefix=env_var_prefix)
        self._validator = None

        if self._file_handler.config_schema:
            self._validator = Validator(
                self._file_handler.config_schema.schema,
                schema_registry=self._file_handler.config_schema.schema_registry,
                allow_unknown=True
            )

    @property
    def config(self) -> CascaDict:
        """ The configuration dictionary. """
        return self._file_handler.configuration

    @config.setter
    def config(self, value: CascaDict):
        self._file_handler.configuration = value

    def validate(self):
        """ Validate the configuration.
        :return: None.
        :raises: :class:`onacol.ConfigValidationError` if configuration
                 is not valid.
        """
        if self._validator is None:
            return

        res = self._validator.validate(self.config)

        if res is False:
            raise ConfigValidationError(
                f"Invalid configuration: {self._validator.errors}")

    def _generate_config_file(self, config: dict, output_file: TextIO) -> None:
        self._file_handler.save_with_schema(config, output_file)

    def generate_config_example(self, output_file: TextIO) -> None:
        self._generate_config_file(self._file_handler.default_config,
                                   output_file)

    def export_current_config(self, output_file: TextIO) -> None:
        self._generate_config_file(self.config, output_file)

    def get_cli_opt_conf_value(self, cli_opt_name: str) -> None:
        return self._flat_schema_handler.get_config_from_cli_opt(
            self.config, cli_opt_name
        )

    def set_cli_opt_conf_value(self, cli_opt_name: str, value: Any) -> Any:
        self._flat_schema_handler.set_config_from_cli_opt(
            self.config, cli_opt_name, value
        )

    def get_env_var_conf_value(self, env_var_name: str) -> Any:
        return self._flat_schema_handler.get_config_from_env_var(
            self.config, env_var_name
        )

    def set_env_var_conf_value(self, env_var_name: str, value: Any) -> None:
        return self._flat_schema_handler.set_config_from_env_var(
            self.config, env_var_name, value
        )

    def merge_env_vars(self, env_var_list: list) -> None:
        """ Merge environment variables from the list with the current
            configuration.
            (Creates new layer in the layered config).

        :param env_var_list:  List of tuples (env_var_name, env_var_value).
        """
        self.config = self.config.cascade()
        for env_var_name, value in env_var_list:
            self.set_env_var_conf_value(env_var_name, value)

    def merge_cli_opts(self, cli_opt_list: list) -> None:
        """ Merge CLI optional arguments from the list with the
            current configuration.
            (Creates new layer in the layered config).

        :param cli_opt_list: List of tuples (cli_opt_name, cli_opt_value).
        """
        self.config = self.config.cascade()
        for cli_opt_name, value in cli_opt_list:
            self.set_cli_opt_conf_value(cli_opt_name, value)

    def config_from_env_vars(self) -> None:
        """ Parse current system's environment variables, merge those with
            valid prefix to the current configuration.
        """
        env_var_list = [(env_var_name, value) for env_var_name, value
                        in os.environ.items() if
                        self._flat_schema_handler.is_prefixed_env_var(
                            env_var_name)]
        self.merge_env_vars(env_var_list)

    def config_from_cli_opts(self, cli_opt_list: list) -> None:
        """ Parse provided CLI optional argument list, merge those with
            valid names to the current configuration.
        """
        parsed_cli_opt_list = [(cli_opt_name, value) for cli_opt_name, value
                        in cli_opt_list if
                        self._flat_schema_handler.is_valid_cli_opt(cli_opt_name)
                               ]
        self.merge_cli_opts(parsed_cli_opt_list)

    def config_from_file(self, file_path: str) -> None:
        """ Load configuration from additional file.
            Configuration will be merged on top of the default/existing config.

        :param file_path: Configuration file path.
        """
        self._file_handler.load_additional_file(file_path)

    def config_from_dict(self, config_dict: dict) -> None:
        """ Load configuration from a dictionary.
            Configuration will be merged on top of the default/existing config.

        :param config_dict:  Configuration dict.
        """
        self.config = self.config.cascade(config_dict)



"""
.. module: onacol.env_var
   :synopsis: Processing for bles.

.. moduleauthor:: Josef Nevrly <josef.nevrly@gmail.com>
"""
from functools import reduce
from typing import Any, NamedTuple
from enum import Enum
from collections import namedtuple
import operator
import json

from cerberus import Validator  # type: ignore

from .base import OnacolException


class UnknownConfigError(OnacolException):
    pass


class InvalidValueError(OnacolException):
    pass


class FlatValueType(Enum):
    VALUE = "val"
    LIST = "list"


FlatSchemaMetadata = namedtuple("FlatSchemaMetadata", "value_type data_type")


class FlatSchemaHandler:

    ENV_VAR_SEPARATOR_CHAR = "_"
    CLI_OPT_SEPARATOR_CHAR = "-"
    SEPARATOR = 2*ENV_VAR_SEPARATOR_CHAR

    def __init__(self, flat_schema: dict, env_var_prefix: str = ""):
        self._flat_schema = flat_schema
        self._prefix = env_var_prefix.lstrip(
            self.ENV_VAR_SEPARATOR_CHAR).upper() + self.ENV_VAR_SEPARATOR_CHAR

        # Initialize env_var identifier mapping from the flat_schema
        # The point of this mapping is, that there should be no limits on
        # how is the configuration schema defined, including config elements
        # that contain separator strings. With this limit, there is no way
        # how to make a two-way mapping from flat schema to real schema
        # and back. So, we need to always have one-way mapping from flat schema
        # to the real schema for both
        #   * env_vars (that are capitalized with prefix)
        #   * cli options (that may or may not be lowercase and have no prefix)
        self._env_var_mapping = {self._prefix +
                                 self.SEPARATOR.join(path).upper(): path
                                 for path, v in self._flat_schema.items()}
        self._cli_opt_mapping = {
            self.SEPARATOR.join(path).replace(self.ENV_VAR_SEPARATOR_CHAR,
                                              self.CLI_OPT_SEPARATOR_CHAR
                                              ): path
            for path, v in self._flat_schema.items()}

    @staticmethod
    def _get_config_value(config, config_path):
        return reduce(operator.getitem, config_path, config)

    def _get_mapped_path(self, mapping, path):
        try:
            return mapping[path]
        except KeyError:
            raise UnknownConfigError(
                f"No configuration exist for {path}")

    def _set_mapped_value(self, config, mapping, path, value):
        mapped_path = self._get_mapped_path(mapping, path)
        metadata = self._flat_schema[mapped_path]

        # Check whether the mapped path points to a value or list
        # If it's from the list, we assume the contents is a JSON encoded list
        if metadata.value_type == FlatValueType.LIST:
            try:
                converted_value = json.loads(value)
                if not isinstance(converted_value, list):
                    raise InvalidValueError(
                        f"Only values of type list expected for {path}")
            except json.JSONDecodeError as e:
                raise InvalidValueError(
                    f"Value {value} is not valid JSON: {str(e)}")
        else:
            # For values, we may try to coerce type conversion, because all
            # values may be string
            if isinstance(value, str) and metadata.data_type is not None:
                # Get possible conversion types from the Cerberos
                type_def = Validator.types_mapping[metadata.data_type]
                # try to convert value from string
                for val_type in type_def.included_types:
                    if isinstance(val_type, tuple):
                        # Somehow, the Cerberus types_mapping wraps int
                        # into an additional tuple...
                        val_type = val_type[0]
                    try:
                        converted_value = val_type(value)
                        break
                    except ValueError:
                        pass
                else:
                    converted_value = value
            else:
                converted_value = value

        self._get_config_value(config, mapped_path[:-1])[mapped_path[-1]] = \
            converted_value

    def _get_config_path_env_var(self, env_var_name):
        return self._get_mapped_path(self._env_var_mapping, env_var_name)

    def _get_config_path_cli_opt(self, cli_opt_name):
        return self._get_mapped_path(self._cli_opt_mapping, cli_opt_name)

    def get_config_from_env_var(self, config: dict, env_var_name: str) -> Any:
        return self._get_config_value(
            config, self._get_config_path_env_var(env_var_name))

    def get_config_from_cli_opt(self, config: dict, cli_opt_name: str) -> Any:
        return self._get_config_value(
            config, self._get_config_path_cli_opt(cli_opt_name))

    def set_config_from_env_var(self, config: dict,
                                env_var_name: str,
                                value: Any) -> None:
        """ Sets value in configuration provided as environment variable.

        :param config:  The configuration dict.
        :param env_var_name:  Environment variable name.
        :param value:  Environment variable value.
        """
        self._set_mapped_value(
            config, self._env_var_mapping, env_var_name, value)

    def set_config_from_cli_opt(self, config: dict,
                                cli_opt_name: str,
                                value: Any) -> None:
        """ Sets value in configuration provided as CLI optional argument.

        :param config:  The configuration dict.
        :param cli_opt_name:  CLI optional argument name.
        :param value:  CLI optional argument value.
        """
        self._set_mapped_value(
            config, self._cli_opt_mapping, cli_opt_name, value)

    def is_prefixed_env_var(self, env_var_name: str) -> bool:
        """ Checks if given environment variable has the specified prefix.

        :param env_var_name: Name of the environment variable.
        :return: True if env_var has specified prefix, False otherwise.
        """
        return env_var_name.startswith(self._prefix)

    def is_valid_cli_opt(self, cli_opt_name: str) -> bool:
        """ Check if CLI optional argument is a valid configuration name.

        :param cli_opt_name: CLI optional argument name.
        :return: True if valid, False otherwise.
        """
        try:
            self._get_config_path_cli_opt(cli_opt_name)
            return True
        except UnknownConfigError:
            return False




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['onacol']

package_data = \
{'': ['*']}

install_requires = \
['Cerberus>=1.3.4,<2.0.0',
 'cascadict>=0.8.4,<0.9.0',
 'ruamel.yaml>=0.17.10,<0.18.0']

setup_kwargs = {
    'name': 'onacol',
    'version': '0.2.3',
    'description': 'Oh No! Another Configuration Library',
    'long_description': '=============================================\nONACOL (Oh No! Another COnfiguration Library)\n=============================================\n\n.. image:: https://badge.fury.io/py/onacol.svg\n        :target: https://badge.fury.io/py/onacol\n\n.. image:: https://github.com/calcite/onacol/actions/workflows/test.yaml/badge.svg?branch=main\n        :target: https://github.com/calcite/onacol/actions/workflows/test.yaml\n\n.. image:: https://readthedocs.org/projects/onacol/badge/?version=latest\n        :target: https://onacol.readthedocs.io/en/latest/?version=latest\n        :alt: Documentation Status\n\n.. image:: https://coveralls.io/repos/github/calcite/onacol/badge.svg?branch=main\n        :target: https://coveralls.io/github/calcite/onacol?branch=main\n        :alt: Test coverage Status\n\n.. image:: https://img.shields.io/pypi/pyversions/onacol\n        :alt: PyPI - Python Version\n\nOnacol is a low-opinionated configuration management library with following\nfeatures:\n\n* YAML (=structured and hierarchical) configuration file support\n* Environment variables support\n* CLI arguments support\n* Configuration merging/overwriting/layering\n* Parameter validation (via Cerberus_)\n* Configuration schema, documentation and default values are defined in\n  single YAML -> No code schema.\n* Minimal dependencies\n\nComparison with other Python configuration libraries/frameworks\n---------------------------------------------------------------\n\nAs the library name suggests, author is painfully aware this is not a unique\nsolution to the problem of application configuration. However, in the plethora\nof existing solutions, none was completely fulfilling the features/requirements\nmentioned above. So, with great reluctance,\n`I had to make my own <https://xkcd.com/927/>`_.\n\nFollowing table lists known/popular configuration frameworks and their\nfeatures relative to Onacol, but not comparing other features that some of those\nlibraries have and Onacol doesn\'t, so check them out - you may find it suits\nyour need better.\n\n\n.. list-table:: Popular configuration framework comparison\n    :widths: 30 10 10 10 10 10 10\n    :header-rows: 1\n\n    * - Framework\n      - YAML\n      - ENV vars\n      - CLI args\n      - Merging\n      - Validation\n      - No code schema\n    * - Hydra_\n      - ✔️\n      - ✔️\n      - ❓\n      - ✔️\n      - ✔️\n      - ✖️\n    * - Pydantic_\n      - ❓\n      - ❓\n      - ✔️\n      - ✔️\n      - ✔️\n      - ✖️\n    * - Dynaconf_\n      - ✔️\n      - ❓\n      - ✔️\n      - ✔️\n      - ✔️\n      - ✖️\n    * - python-dotenv_\n      - ✖️\n      - ✔️\n      - ✖️\n      - ✖️\n      - ✖️\n      - ✖️\n    * - `Gin Config`_\n      - ❓\n      - ❓\n      - ❓\n      - ❓\n      - ✔️\n      - ✖️\n    * - OmegaConf_\n      - ✔️\n      - ✔️\n      - ✔️\n      - ✔️\n      - ✔️\n      - ✖️\n    * - Confuse_\n      - ✔️\n      - ✔️\n      - ❓\n      - ✔️\n      - ✔️\n      - ✖️\n    * - `Python Decouple`_\n      - ✖️\n      - ✖️\n      - ✔️\n      - ✔️\n      - ✖️\n      - ✖️\n    * - parse_it_\n      - ✔️\n      - ✔️\n      - ✔️\n      - ✔️\n      - ❓\n      - ✖️\n    * - Grift_\n      - ✖️\n      - ✖️\n      - ✖️\n      - ❓\n      - ✔️\n      - ✖️\n    * - profig_\n      - ✖️\n      - ✔️\n      - ✖️\n      - ❓\n      - ✔️\n      - ✖️\n    * - tweak_\n      - ✔️\n      - ✖️\n      - ✖️\n      - ✔️\n      - ✖️\n      - ✖️\n    * - Bison_\n      - ✔️\n      - ❓\n      - ✔️\n      - ✔️\n      - ✔️\n      - ✖️\n    * - Config-Man_\n      - ✖️\n      - ✔️\n      - ✔️\n      - ❓\n      - ✔️\n      - ✖️\n    * - figga_\n      - ✔️\n      - ✖️\n      - ✔️\n      - ❓\n      - ✖️\n      - ✖️\n    * - **Onacol**\n      - ✔️\n      - ✔️\n      - ✔️\n      - ✔️\n      - ✔️\n      - ✔️\n\nInstallation\n------------\n\nAs usually with pip::\n\n    $ pip install onacol\n\nUsage\n-----\n\nDefault configuration file & schema\n+++++++++++++++++++++++++++++++++++\n\nThe whole point of this library is the definition of both default configuration\nand configuration schema in one YAML file (i.e. single source of configuration\ntruth).\n\nLet\'s start with a simple `default_config.yaml` file that is part of an\napplication\'s package. This file contains default values for the configuration.\n\n.. code-block:: yaml\n\n    general:\n        # Logging level for this application.\n        log_level: INFO\n\n    ui:\n        # Address and port of the UI webserver\n        addr: 0.0.0.0\n        port: 8888\n\n    sensors:\n        sensor_reset_interval: 30.0  # Sensor reset interval in seconds\n        connected_units:\n            - id: 0                     # Sensor ID <0, 16>\n              name: "Basic sensor"\n              min_trigger_limit: 30     # Minimal triggering limit [cm]\n              max_trigger_limit: 120    # Maximal triggering limit [cm]\n            - id: 1\n              name: "Additional sensor"\n              min_trigger_limit: 40\n              max_trigger_limit: 100\n\nThis file can be used as it is. However, we can add a schema definition to the\nstructure, that will allow parameter validation and automatic type conversion.\n\nThis is done by adding metadata to the YAML structure. Following metadata are\nrecognized by Onacol:\n\n* `oc_schema`: Cerberus_ validator/schema definitions.\n* `oc_default`: Default value (if metadata are attached to the YAML element, it\n  can no longer bear the value directly.\n* `oc_schema_id`: Definition of a schema reference (see TODO)\n\nSchema metadata are NOT MANDATORY. We can only provide them to parameters for\nwhich we think validation (or type conversion) may be useful.\n\n.. code-block:: yaml\n\n    general:\n        # Logging level for this application.\n        log_level: INFO\n\n    ui:\n        # Address and port of the UI webserver\n        addr:\n            oc_default: 0.0.0.0\n            oc_schema:\n                type: string\n                regex: "^(?:[0-9]{1,3}\\.){3}[0-9]{1,3}$"\n\n        port:\n            oc_default: 8888\n            oc_schema:\n                type: integer\n\n    sensors:\n        sensor_reset_interval:          # Sensor reset interval in seconds\n            oc_default: 30.0\n            oc_schema:\n                type: float\n                min: 0.0\n                max: 100.0\n        connected_units:\n            - id:                       # Sensor ID <0, 16>\n                oc_default: 0\n                oc_schema:\n                    type: integer\n                    min: 0\n                    max: 16\n              name: "Basic sensor"\n              min_trigger_limit:        # Minimal triggering limit [cm]\n                oc_default: 30\n                oc_schema:\n                    type: integer\n                    min: 0\n                    max: 200\n              max_trigger_limit:        # Maximal triggering limit [cm]\n                oc_default: 120\n                oc_schema:\n                    type: integer\n                    min: 0\n                    max: 200\n            - id: 1\n              name: "Additional sensor"\n              min_trigger_limit: 40\n              max_trigger_limit: 100\n\nNote that for list definitions, schema is added only to the first element of the\nlist. Other elements will be validated based on the first element\'s schema.\n\n\nLoading and validating configuration in an application\n++++++++++++++++++++++++++++++++++++++++++++++++++++++\n\nOnacol is used by the application via the `ConfigManager` instance.\n`ConfigManager` can load configurations from multiple sources (files,\ncommand line optional arguments, environment variables), but does not do it\nautomatically - the sources and order is up to the app implementation.\n\nExample (using Click_ as a CLI framework):\nTODO\n\n\n* Free software: MIT license\n* Documentation: https://onacol.readthedocs.io.\n\n\nFeatures\n--------\n\n* TODO\n\nCredits\n-------\n\nThis package was created with Cookiecutter_ and the `audreyr/cookiecutter-pypackage`_ project template.\n\n\nLimitations\n-----------\n\nVariable-count structures must be contained in lists.\nComments following oc_* tags are not kept.\n\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _`audreyr/cookiecutter-pypackage`: https://github.com/audreyr/cookiecutter-pypackage\n.. _Cerberus: https://docs.python-cerberus.org/en/stable/\n.. _Hydra: https://hydra.cc/\n.. _Config-Man: https://github.com/mmohaveri/config-man\n.. _Dynaconf: https://github.com/rochacbruno/dynaconf\n.. _Pydantic: https://pydantic-docs.helpmanual.io/\n.. _python-dotenv: https://github.com/theskumar/python-dotenv\n.. _`Gin Config`: https://github.com/google/gin-config\n.. _OmegaConf: https://github.com/omry/omegaconf\n.. _Confuse: https://github.com/beetbox/confuse\n.. _`Python Decouple`: https://github.com/henriquebastos/python-decouple\n.. _parse_it: https://github.com/naorlivne/parse_it\n.. _grift: https://github.com/kensho-technologies/grift\n.. _profig: https://github.com/dhagrow/profig\n.. _tweak: https://github.com/kislyuk/tweak\n.. _Bison: https://github.com/edaniszewski/bison\n.. _figga: https://github.com/berislavlopac/figga\n.. _Click: https://click.palletsprojects.com\n',
    'author': 'Josef Nevrly',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
